"""Optional functions for uploading data direct to S3."""
import contextlib
from io import BytesIO, TextIOWrapper
from tempfile import TemporaryFile
from typing import IO, Generator, Union

try:
    import boto3
except ImportError:
    raise ImportError("You cannot use the django_csv.s3 module without boto3.")

# type used to smooth over TemporaryFile <> BytesIO mismatch
FileLikeObject = Union[IO[bytes], BytesIO]


# extracted out to facilitate testing
def _put_object(bucket: str, key: str, fileobj: FileLikeObject) -> None:
    """Upload binary stream to S3 using put_pubject."""
    client = boto3.client("s3")
    client.put_object(Bucket=bucket, Key=key, Body=fileobj, ContentType="text/csv")


# extracted out to facilitate testing
def _upload_fileobj(bucket: str, key: str, fileobj: FileLikeObject) -> None:
    """Upload binary stream to S3 using upload_fileobj."""
    client = boto3.client("s3")
    client.upload_fileobj(fileobj, bucket, key)


@contextlib.contextmanager
def s3_upload_fileobj(bucket: str, key: str) -> Generator:
    """
    Context manager used to write to S3 using upload_fileobj.

    This context manager writes to a TemporaryFile and then uses the
    multipart boto3 upload function `upload_fileobj`, and is more
    appropriate for large files.

    >>> with s3_upload_fileobj("bucket", "obj_key") as fileobj:
    ...     write_csv(fileobj, queryset, "col1", "col2")

    """
    with TemporaryFile() as fileobj:
        with TextIOWrapper(
            fileobj, encoding="utf-8", newline="", write_through=True
        ) as buffer:
            yield buffer
            fileobj.seek(0)
            _upload_fileobj(bucket, key, fileobj)


@contextlib.contextmanager
def s3_put_object(bucket: str, key: str) -> Generator:
    """
    Context manager used to write to S3 using put_object.

    This context manager writes to a in-memory buffer and then uses the
    one-shot boto3 upload function `put_object`, and is more appropriate
    for smaller files.

    >>> with s3_put_object("bucket", "obj_key") as fileobj:
    ...     write_csv(fileobj, queryset, "col1", "col2")

    """
    with BytesIO() as fileobj:
        with TextIOWrapper(
            fileobj, encoding="utf-8", newline="", write_through=True
        ) as buffer:
            yield buffer
            fileobj.seek(0)
            _put_object(bucket, key, fileobj)



from typing import Sequence

from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.db.models.query import QuerySet
from django.http import HttpRequest, HttpResponse
from django.views import View

from .csv import write_csv
from .models import CsvDownload


def download_csv(
    user: settings.AUTH_USER_MODEL, filename: str, queryset: QuerySet, *columns: str
) -> HttpResponse:
    """Download queryset as a CSV."""
    response = HttpResponse(content_type="text/csv")
    response["Content-Disposition"] = f'attachment; filename="{filename}"'
    row_count = write_csv(response, queryset, *columns)
    response["X-Row-Count"] = row_count
    CsvDownload.objects.create(
        user=user,
        row_count=row_count,
        filename=filename,
        columns=", ".join(columns),
    )
    return response


class CsvDownloadView(View):
    """CBV for downloading CSVs."""

    def has_permission(self, request: HttpRequest) -> bool:
        """Return True if the user has permission to download this file."""
        return True

    def get_user(self, request: HttpRequest) -> settings.AUTH_USER_MODEL:
        """
        Return the user against whom to record the download.

        This is provided for cases where the request.user may not be the
        user you wish to record the download against. Impersonation is the
        canonical use case for overriding this.

        """
        return request.user

    def get_filename(self, request: HttpRequest) -> str:
        """Return download filename."""
        raise NotImplementedError

    def get_columns(self, request: HttpRequest) -> Sequence[str]:
        """Return columns to extract from the queryset."""
        raise NotImplementedError

    def get_queryset(self, request: HttpRequest) -> QuerySet:
        """Return the data to be downloaded."""
        raise NotImplementedError

    def get(self, request: HttpRequest) -> HttpResponse:
        """Download data as CSV."""
        if not self.has_permission(request):
            raise PermissionDenied

        return download_csv(
            self.get_user(request),
            self.get_filename(request),
            self.get_queryset(request),
            *self.get_columns(request),
        )
